use axum::response::{IntoResponse, Redirect};
use reqwest::StatusCode;

#[cfg(feature = "oidc")]
use axum_extra::extract::PrivateCookieJar;

#[cfg(feature = "oidc")]
use openidconnect::{
    core::CoreErrorResponseType, url::ParseError, ClaimsVerificationError, DiscoveryError,
    SigningError, StandardErrorResponse,
};
#[cfg(feature = "oidc")]
type RequestTokenError = openidconnect::RequestTokenError<
    openidconnect::reqwest::Error<reqwest::Error>,
    StandardErrorResponse<CoreErrorResponseType>,
>;

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[cfg(feature = "oidc")]
    #[error("discovery error: {:?}", 0)]
    Discovery(#[from] DiscoveryError<openidconnect::reqwest::Error<reqwest::Error>>),

    #[cfg(feature = "oidc")]
    #[error("parse error: {:?}", 0)]
    Parse(#[from] ParseError),

    #[cfg(feature = "oidc")]
    #[error("request token error: {:?}", 0)]
    RequestToken(#[from] RequestTokenError),

    #[cfg(feature = "oidc")]
    #[error("claims verification error: {:?}", 0)]
    ClaimsVerification(#[from] ClaimsVerificationError),

    #[cfg(feature = "oidc")]
    #[error("signing error: {:?}", 0)]
    Signing(#[from] SigningError),

    #[error("json serialization error: {:?}", 0)]
    Json(#[from] serde_json::Error),

    #[error("url parsing error: {:?}", 0)]
    UrlParsing(#[from] axum::http::Error),

    #[cfg(feature = "oidc")]
    #[error("csrf token is invalid")]
    CsrfTokenInvalid,

    #[cfg(feature = "oidc")]
    #[error("id token not found")]
    IdTokenNotFound,

    #[cfg(feature = "oidc")]
    #[error("access token hash is invalid")]
    AccessTokenHashInvalid,

    #[cfg(feature = "oidc")]
    #[error("just a redirect")]
    Redirect((PrivateCookieJar, Redirect)),

    #[error("reqwest: {0}")]
    Reqwest(#[from] reqwest::Error),

    #[cfg(feature = "jwt")]
    #[error("jsonwebtoken: {0}")]
    JsonWebToken(#[from] jsonwebtoken::errors::Error),

    #[cfg(feature = "jwt")]
    #[error("jwt invalid")]
    JwtInvalid,
}

impl IntoResponse for Error {
    fn into_response(self) -> axum::response::Response {
        match self {
            #[cfg(feature = "oidc")]
            Self::CsrfTokenInvalid => {
                { (StatusCode::BAD_REQUEST, "csrf token is invalid").into_response() }
                    .into_response()
            }

            #[cfg(feature = "oidc")]
            Self::Redirect(redirect) => redirect.into_response(),

            #[cfg(feature = "jwt")]
            Self::JwtInvalid => (StatusCode::UNAUTHORIZED, "access token invalid").into_response(),

            _ => (StatusCode::INTERNAL_SERVER_ERROR, "internal server error").into_response(),
        }
    }
}
