use std::str::FromStr;

use async_trait::async_trait;
use axum::{
    extract::{FromRef, FromRequestParts, Query},
    http::{request::Parts, Uri},
    response::Redirect,
};
pub use axum_extra::extract::cookie::Key;
use axum_extra::extract::{
    cookie::{Cookie, SameSite},
    PrivateCookieJar,
};
use error::Error;
use openidconnect::{
    core::{
        CoreAuthDisplay, CoreAuthPrompt, CoreAuthenticationFlow, CoreErrorResponseType,
        CoreGenderClaim, CoreJsonWebKey, CoreJsonWebKeyType, CoreJsonWebKeyUse,
        CoreJweContentEncryptionAlgorithm, CoreJwsSigningAlgorithm, CoreProviderMetadata,
        CoreRevocableToken, CoreRevocationErrorResponse, CoreTokenIntrospectionResponse,
        CoreTokenType,
    },
    reqwest::async_http_client,
    AccessTokenHash, AuthorizationCode, Client, ClientId, ClientSecret, CsrfToken,
    EmptyExtraTokenFields, IdTokenClaims, IdTokenFields, IssuerUrl, Nonce, OAuth2TokenResponse,
    PkceCodeChallenge, PkceCodeVerifier, RedirectUrl, Scope, StandardErrorResponse,
    StandardTokenResponse, TokenResponse,
};
use serde::{Deserialize, Serialize};

use crate::error;

const LOGIN_COOKIE_NAME: &str = "OIDC_LOGIN";

pub trait AdditionalClaims: openidconnect::AdditionalClaims + Clone + Sync + Send {}

type OidcTokenResponse<AC> = StandardTokenResponse<
    IdTokenFields<
        AC,
        EmptyExtraTokenFields,
        CoreGenderClaim,
        CoreJweContentEncryptionAlgorithm,
        CoreJwsSigningAlgorithm,
        CoreJsonWebKeyType,
    >,
    CoreTokenType,
>;

pub type OidcClient<AC> = Client<
    AC,
    CoreAuthDisplay,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJwsSigningAlgorithm,
    CoreJsonWebKeyType,
    CoreJsonWebKeyUse,
    CoreJsonWebKey,
    CoreAuthPrompt,
    StandardErrorResponse<CoreErrorResponseType>,
    OidcTokenResponse<AC>,
    CoreTokenType,
    CoreTokenIntrospectionResponse,
    CoreRevocableToken,
    CoreRevocationErrorResponse,
>;

pub type IdToken<AZ> = openidconnect::IdToken<
    AZ,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJwsSigningAlgorithm,
    CoreJsonWebKeyType,
>;

#[derive(Clone)]
pub struct OidcApplication<AC: AdditionalClaims> {
    application_base: Uri,
    scopes: Vec<String>,
    cookie_key: Key,
    client: OidcClient<AC>,
}
impl<AC: AdditionalClaims> OidcApplication<AC> {
    pub async fn create(
        application_base: Uri,
        issuer: String,
        client_id: String,
        client_secret: Option<String>,
        scopes: Vec<String>,
        cookie_key: Key,
    ) -> Result<Self, Error> {
        let provider_metadata = CoreProviderMetadata::discover_async(
            IssuerUrl::new(issuer).unwrap(),
            async_http_client,
        )
        .await?;
        let client = OidcClient::<AC>::from_provider_metadata(
            provider_metadata,
            ClientId::new(client_id),
            client_secret.map(ClientSecret::new),
        );

        Ok(Self {
            application_base,
            scopes,
            cookie_key,
            client,
        })
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct EmptyAdditionalClaims {}
impl openidconnect::AdditionalClaims for EmptyAdditionalClaims {}
impl AdditionalClaims for EmptyAdditionalClaims {}

pub struct OidcExtractor<AC: AdditionalClaims> {
    pub claims: IdTokenClaims<AC, CoreGenderClaim>,
    pub access_token: String,
}

#[async_trait]
impl<S, AC> FromRequestParts<S> for OidcExtractor<AC>
where
    S: Send + Sync,
    AC: AdditionalClaims,
    OidcApplication<AC>: FromRef<S>,
{
    type Rejection = Error;

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let application: OidcApplication<AC> = OidcApplication::from_ref(state);

        let handler_uri = Uri::builder()
            .scheme(application.application_base.scheme().unwrap().clone())
            .authority(application.application_base.authority().unwrap().clone())
            .path_and_query(strip_oidc_from_path(&parts.uri))
            .build()?;

        let mut client = application.client;
        client = client.set_redirect_uri(RedirectUrl::new(handler_uri.to_string())?);
        let mut jar = PrivateCookieJar::from_headers(&parts.headers, application.cookie_key);

        let login_session = jar.get(LOGIN_COOKIE_NAME);
        let query = Query::<OidcQuery>::from_request_parts(parts, state)
            .await
            .ok();

        if let Some(login_session) = &login_session {
            let login_session: LoginSession = serde_json::from_str(login_session.value())?;
            if let Some(id_token) = login_session.id_token {
                let id_token = IdToken::<AC>::from_str(&id_token).unwrap();
                if let Ok(claims) =
                    id_token.claims(&client.id_token_verifier(), &login_session.nonce)
                {
                    return Ok(Self {
                        claims: claims.clone(),
                        access_token: login_session.access_token.unwrap_or_default(),
                    });
                }
            }
        }

        if let (Some(login_session), Some(Query(query))) = (login_session, query) {
            let mut login_session: LoginSession = serde_json::from_str(login_session.value())?;

            if login_session.csrf_token.secret() != &query.state {
                return Err(Error::CsrfTokenInvalid);
            }

            let token_response = client
                .exchange_code(AuthorizationCode::new(query.code.to_string()))
                // Set the PKCE code verifier.
                .set_pkce_verifier(PkceCodeVerifier::new(
                    login_session.pkce_verifier.secret().to_string(),
                ))
                .request_async(async_http_client)
                .await?;

            // Extract the ID token claims after verifying its authenticity and nonce.
            let id_token = token_response.id_token().ok_or(Error::IdTokenNotFound)?;
            let claims = id_token.claims(&client.id_token_verifier(), &login_session.nonce)?;

            // Verify the access token hash to ensure that the access token hasn't been substituted for
            // another user's.
            if let Some(expected_access_token_hash) = claims.access_token_hash() {
                let actual_access_token_hash = AccessTokenHash::from_token(
                    token_response.access_token(),
                    &id_token.signing_alg()?,
                )?;
                if actual_access_token_hash != *expected_access_token_hash {
                    return Err(Error::AccessTokenHashInvalid);
                }
            }

            login_session.id_token = Some(id_token.to_string());
            login_session.access_token = Some(token_response.access_token().secret().to_string());

            let login_session = serde_json::to_string(&login_session)?;
            jar = jar.add(create_cookie(login_session));

            Err(Error::Redirect((
                jar,
                Redirect::temporary(
                    handler_uri
                        .path_and_query()
                        .map(|x| x.as_str())
                        .unwrap_or(handler_uri.path()),
                ),
            )))
        } else {
            let (pkce_challenge, pkce_verifier) = PkceCodeChallenge::new_random_sha256();
            let (auth_url, csrf_token, nonce) = {
                let mut auth = client.authorize_url(
                    CoreAuthenticationFlow::AuthorizationCode,
                    CsrfToken::new_random,
                    Nonce::new_random,
                );

                for scope in application.scopes.iter() {
                    auth = auth.add_scope(Scope::new(scope.to_string()));
                }

                auth.set_pkce_challenge(pkce_challenge).url()
            };

            let login_session = LoginSession {
                nonce,
                csrf_token,
                pkce_verifier,
                id_token: None,
                access_token: None,
            };
            let login_session = serde_json::to_string(&login_session)?;
            jar = jar.add(create_cookie(login_session));

            Err(Error::Redirect((
                jar,
                Redirect::temporary(auth_url.as_str()),
            )))
        }
    }
}

fn create_cookie(login_session: String) -> Cookie<'static> {
    let mut cookie = Cookie::new(LOGIN_COOKIE_NAME, login_session);
    cookie.set_same_site(SameSite::Lax);
    cookie.set_secure(true);
    cookie.set_http_only(true);
    cookie
}

fn strip_oidc_from_path(uri: &Uri) -> String {
    let query = uri
        .query()
        .map(|uri| {
            uri.split('&')
                .filter(|x| {
                    !x.starts_with("code")
                        && !x.starts_with("state")
                        && !x.starts_with("session_state")
                })
                .fold(String::new(), |acc, x| acc + "&" + x)
                .chars()
                .skip(1)
                .collect::<String>()
        })
        .unwrap_or_default();
    uri.path().to_string() + &query
}

#[derive(Debug, Deserialize)]
struct OidcQuery {
    code: String,
    state: String,
    #[allow(dead_code)]
    session_state: String,
}

#[derive(Serialize, Deserialize)]
struct LoginSession {
    nonce: Nonce,
    csrf_token: CsrfToken,
    pkce_verifier: PkceCodeVerifier,
    id_token: Option<String>,
    access_token: Option<String>,
}
